#!/usr/bin/env python
#
# Command line tool to take in ESP-IDF sdkconfig files with project
# settings and output data in multiple formats (update config, generate
# header file, generate .cmake include file, documentation, etc).
#
# Used internally by the ESP-IDF build system. But designed to be
# non-IDF-specific.
#
# SPDX-FileCopyrightText: 2018-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import argparse
import json
import os
import os.path
import re
import sys
import tempfile
import textwrap
from collections import OrderedDict, defaultdict

import kconfiglib

import esp_idf_kconfig.gen_kconfig_doc as gen_kconfig_doc
from esp_idf_kconfig import __version__


class DeprecatedOptions(object):
    _REN_FILE = "sdkconfig.rename"
    _DEP_OP_BEGIN = "# Deprecated options for backward compatibility"
    _DEP_OP_END = "# End of deprecated options"
    _RE_DEP_OP_BEGIN = re.compile(_DEP_OP_BEGIN)
    _RE_DEP_OP_END = re.compile(_DEP_OP_END)

    def __init__(self, config_prefix, path_rename_files=[]):
        self.config_prefix = config_prefix
        # r_dic maps deprecated options to new options; rev_r_dic maps in the opposite direction
        self.r_dic, self.rev_r_dic = self._parse_replacements(path_rename_files)

        # note the '=' at the end of regex for not getting partial match of configs.
        # Also match if the config option is followed by a whitespace, this is the case
        # in sdkconfig.defaults files contaning "# CONFIG_MMM_NNN is not set".
        self._RE_CONFIG = re.compile(r"{}(\w+)(=|\s+)".format(self.config_prefix))

    def _parse_replacements(self, repl_paths):
        rep_dic = {}
        rev_rep_dic = defaultdict(list)

        def remove_config_prefix(string):
            if string.startswith(self.config_prefix):
                return string[len(self.config_prefix) :]
            raise RuntimeError(
                "Error in {} (line {}): Config {} is not prefixed with {}"
                "".format(rep_path, line_number, string, self.config_prefix)
            )

        for rep_path in repl_paths:
            with open(rep_path) as f_rep:
                for line_number, line in enumerate(f_rep, start=1):
                    sp_line = line.split()
                    if len(sp_line) == 0 or sp_line[0].startswith("#"):
                        # empty line or comment
                        continue
                    if len(sp_line) != 2 or not all(
                        x.startswith(self.config_prefix) for x in sp_line
                    ):
                        raise RuntimeError(
                            "Syntax error in {} (line {})".format(rep_path, line_number)
                        )
                    if sp_line[0] in rep_dic:
                        raise RuntimeError(
                            "Error in {} (line {}): Replacement {} exist for {} and new "
                            "replacement {} is defined".format(
                                rep_path,
                                line_number,
                                rep_dic[sp_line[0]],
                                sp_line[0],
                                sp_line[1],
                            )
                        )

                    (dep_opt, new_opt) = (remove_config_prefix(x) for x in sp_line)
                    rep_dic[dep_opt] = new_opt
                    rev_rep_dic[new_opt].append(dep_opt)
        return rep_dic, rev_rep_dic

    def get_deprecated_option(self, new_option):
        return self.rev_r_dic.get(new_option, [])

    def get_new_option(self, deprecated_option):
        return self.r_dic.get(deprecated_option, None)

    def replace(self, sdkconfig_in, sdkconfig_out):
        replace_enabled = True
        with open(sdkconfig_in, "r") as f_in, open(sdkconfig_out, "w") as f_out:
            for line_num, line in enumerate(f_in, start=1):
                if self._RE_DEP_OP_BEGIN.search(line):
                    replace_enabled = False
                elif self._RE_DEP_OP_END.search(line):
                    replace_enabled = True
                elif replace_enabled:
                    m = self._RE_CONFIG.search(line)
                    if m and m.group(1) in self.r_dic:
                        depr_opt = self.config_prefix + m.group(1)
                        new_opt = self.config_prefix + self.r_dic[m.group(1)]
                        line = line.replace(depr_opt, new_opt)
                        print(
                            "{}:{} {} was replaced with {}".format(
                                sdkconfig_in, line_num, depr_opt, new_opt
                            )
                        )
                f_out.write(line)

    def append_doc(self, config, visibility, path_output):
        def option_was_written(opt):
            # named choices were written if any of the symbols in the choice were visible
            if new_opt in config.named_choices:
                syms = config.named_choices[new_opt].syms
                for s in syms:
                    if any(visibility.visible(node) for node in s.nodes):
                        return True
                return False
            else:
                try:
                    # otherwise if any of the nodes associated with the option was visible
                    return any(
                        visibility.visible(node) for node in config.syms[opt].nodes
                    )
                except KeyError:
                    return False

        if len(self.r_dic) > 0:
            with open(path_output, "a") as f_o:
                header = "Deprecated options and their replacements"
                f_o.write(
                    ".. _configuration-deprecated-options:\n\n{}\n{}\n\n".format(
                        header, "-" * len(header)
                    )
                )
                for dep_opt in sorted(self.r_dic):
                    new_opt = self.r_dic[dep_opt]
                    if option_was_written(new_opt) and (
                        new_opt not in config.syms
                        or config.syms[new_opt].choice is None
                    ):
                        # everything except config for a choice (no link reference for those in the docs)
                        f_o.write(
                            "- {}{} (:ref:`{}{}`)\n".format(
                                config.config_prefix,
                                dep_opt,
                                config.config_prefix,
                                new_opt,
                            )
                        )

                        if new_opt in config.named_choices:
                            # here are printed config options which were filtered out
                            syms = config.named_choices[new_opt].syms
                            for sym in syms:
                                if sym.name in self.rev_r_dic:
                                    # only if the symbol has been renamed
                                    dep_names = self.rev_r_dic[sym.name]
                                    dep_names = [
                                        config.config_prefix + name
                                        for name in dep_names
                                    ]
                                    # config options doesn't have references
                                    f_o.write("    - {}\n".format(", ".join(dep_names)))

    def append_config(self, config, path_output):
        tmp_list = []

        def append_config_node_process(node):
            item = node.item
            if isinstance(item, kconfiglib.Symbol) and item.env_var is None:
                if item.name in self.rev_r_dic:
                    c_string = item.config_string
                    if c_string:
                        for dep_name in self.rev_r_dic[item.name]:
                            tmp_list.append(
                                c_string.replace(
                                    self.config_prefix + item.name,
                                    self.config_prefix + dep_name,
                                )
                            )

        for n in config.node_iter():
            append_config_node_process(n)

        if len(tmp_list) > 0:
            with open(path_output, "a") as f_o:
                f_o.write("\n{}\n".format(self._DEP_OP_BEGIN))
                f_o.writelines(tmp_list)
                f_o.write("{}\n".format(self._DEP_OP_END))

    def append_header(self, config, path_output):
        def _opt_defined(opt):
            if (
                opt.orig_type in (kconfiglib.BOOL, kconfiglib.TRISTATE)
                and opt.str_value != "n"
            ):
                opt_defined = True
            elif (
                opt.orig_type in (kconfiglib.INT, kconfiglib.STRING, kconfiglib.HEX)
                and opt.str_value != ""
            ):
                opt_defined = True
            else:
                opt_defined = False
            return opt_defined

        if len(self.r_dic) > 0:
            with open(path_output, "a") as f_o:
                f_o.write("\n/* List of deprecated options */\n")
                for dep_opt in sorted(self.r_dic):
                    new_opt = self.r_dic[dep_opt]
                    if new_opt in config.syms and _opt_defined(config.syms[new_opt]):
                        f_o.write(
                            "#define {}{} {}{}\n".format(
                                self.config_prefix, dep_opt, self.config_prefix, new_opt
                            )
                        )


def main():
    parser = argparse.ArgumentParser(
        description="kconfgen.py v%s - Config Generation Tool" % __version__,
        prog=os.path.basename(sys.argv[0]),
    )

    parser.add_argument(
        "--config", help="Project configuration settings", nargs="?", default=None
    )

    parser.add_argument(
        "--defaults",
        help="Optional project defaults file, used if --config file doesn't exist. "
        "Multiple files can be specified using multiple --defaults arguments.",
        nargs="?",
        default=[],
        action="append",
    )

    parser.add_argument(
        "--kconfig", help="KConfig file with config item definitions", required=True
    )

    parser.add_argument(
        "--sdkconfig-rename",
        help="File with deprecated Kconfig options",
        required=False,
    )

    parser.add_argument(
        "--dont-write-deprecated",
        help="Do not write compatibility statements for deprecated values",
        action="store_true",
    )

    parser.add_argument(
        "--output",
        nargs=2,
        action="append",
        help="Write output file (format and output filename)",
        metavar=("FORMAT", "FILENAME"),
        default=[],
    )

    parser.add_argument(
        "--env",
        action="append",
        default=[],
        help="Environment to set when evaluating the config file",
        metavar="NAME=VAL",
    )

    parser.add_argument(
        "--env-file",
        type=argparse.FileType("r"),
        help="Optional file to load environment variables from. Contents "
        "should be a JSON object where each key/value pair is a variable.",
    )

    parser.add_argument(
        "--list-separator",
        choices=["space", "semicolon"],
        default="space",
        help="Separator used in environment list variables (COMPONENT_SDKCONFIG_RENAMES)",
    )

    args = parser.parse_args()

    for fmt, filename in args.output:
        if fmt not in OUTPUT_FORMATS.keys():
            print(
                "Format '%s' not recognised. Known formats: %s"
                % (fmt, OUTPUT_FORMATS.keys())
            )
            sys.exit(1)

    try:
        args.env = [
            (name, value) for (name, value) in (e.split("=", 1) for e in args.env)
        ]
    except ValueError:
        print(
            "--env arguments must each contain =. To unset an environment variable, use 'ENV='"
        )
        sys.exit(1)

    for name, value in args.env:
        os.environ[name] = value

    if args.env_file is not None:
        env = json.load(args.env_file)
        os.environ.update(env)

    config = kconfiglib.Kconfig(args.kconfig)
    config.warn_assign_redun = False
    config.warn_assign_override = False

    sdkconfig_renames_sep = ";" if args.list_separator == "semicolon" else " "

    sdkconfig_renames = [args.sdkconfig_rename] if args.sdkconfig_rename else []
    sdkconfig_renames_from_env = os.environ.get("COMPONENT_SDKCONFIG_RENAMES")
    if sdkconfig_renames_from_env:
        sdkconfig_renames += sdkconfig_renames_from_env.split(sdkconfig_renames_sep)
    deprecated_options = DeprecatedOptions(
        config.config_prefix, path_rename_files=sdkconfig_renames
    )

    if len(args.defaults) > 0:

        def _replace_empty_assignments(path_in, path_out):
            with open(path_in, "r") as f_in, open(path_out, "w") as f_out:
                for line_num, line in enumerate(f_in, start=1):
                    line = line.strip()
                    if line.endswith("="):
                        line += "n"
                        print(
                            "{}:{} line was updated to {}".format(
                                path_out, line_num, line
                            )
                        )
                    f_out.write(line)
                    f_out.write("\n")

        # always load defaults first, so any items which are not defined in that config
        # will have the default defined in the defaults file
        for name in args.defaults:
            print("Loading defaults file %s..." % name)
            if not os.path.exists(name):
                raise RuntimeError("Defaults file not found: %s" % name)
            try:
                with tempfile.NamedTemporaryFile(
                    prefix="kconfgen_tmp", delete=False
                ) as f:
                    temp_file1 = f.name
                with tempfile.NamedTemporaryFile(
                    prefix="kconfgen_tmp", delete=False
                ) as f:
                    temp_file2 = f.name
                deprecated_options.replace(sdkconfig_in=name, sdkconfig_out=temp_file1)
                _replace_empty_assignments(temp_file1, temp_file2)
                config.load_config(temp_file2, replace=False)

                for symbol, value in config.missing_syms:
                    if deprecated_options.get_new_option(symbol) is None:
                        print(
                            f"warning: unknown kconfig symbol '{symbol}' assigned to '{value}' in {name}"
                        )
            finally:
                try:
                    os.remove(temp_file1)
                    os.remove(temp_file2)
                except OSError:
                    pass

    # If config file previously exists, load it
    if args.config and os.path.exists(args.config):
        # ... but replace deprecated options before that
        with tempfile.NamedTemporaryFile(prefix="kconfgen_tmp", delete=False) as f:
            temp_file = f.name
        try:
            deprecated_options.replace(
                sdkconfig_in=args.config, sdkconfig_out=temp_file
            )
            config.load_config(temp_file, replace=False)
            update_if_changed(temp_file, args.config)
        finally:
            try:
                os.remove(temp_file)
            except OSError:
                pass

    if args.dont_write_deprecated:
        # The deprecated object was useful until now for replacements. Now it will be redefined with no configurations
        # and as the consequence, it won't generate output with deprecated statements.
        deprecated_options = DeprecatedOptions("", path_rename_files=[])

    # Output the files specified in the arguments
    for output_type, filename in args.output:
        with tempfile.NamedTemporaryFile(prefix="kconfgen_tmp", delete=False) as f:
            temp_file = f.name
        try:
            output_function = OUTPUT_FORMATS[output_type]
            output_function(deprecated_options, config, temp_file)
            update_if_changed(temp_file, filename)
        finally:
            try:
                os.remove(temp_file)
            except OSError:
                pass


def write_config(deprecated_options, config, filename):
    idf_version = os.environ.get("IDF_VERSION", "")
    CONFIG_HEADING = f"""#
# Automatically generated file. DO NOT EDIT.
# Espressif IoT Development Framework (ESP-IDF) {idf_version} Project Configuration
#
"""
    config.write_config(filename, header=CONFIG_HEADING)
    deprecated_options.append_config(config, filename)


def min_config_with_labels(config: kconfiglib.Kconfig, header: str) -> str:
    """Return minimal config containing menu labels"""
    all_options = config._config_contents("").splitlines()
    min_options = config._min_config_contents("").splitlines()

    end_regex = re.compile(r"^# end of (.*)$").match
    # `label_path` marks the current path from the root; labels represent tree nodes.
    # The path is represented as an ordered Dict; the last element in the Dict is closest to the leaves.
    # The label name is the key, and a boolean value indicates whether the label was added to the output (i.e., the label is used in the output).
    # We need to track used labels to ensure correct timing when printing the label ending.
    # Note that label names are not necessarily unique, so the order is significant.
    label_path: OrderedDict[str, bool] = OrderedDict()
    output = [header]
    possibly_label = False
    current = None  # None stands for tree root
    comments = []

    # Using depth search first, we go down the tree and save the path from the root.
    # If we find an option from min config, we update the whole path to used (True) and print all menu labels.
    # When we go back up the tree, we print all label endings that were marked as used
    for line in all_options:
        end_match = end_regex(line)
        if end_match:
            # we have found an end of a menu section
            current = end_match.group(1)
            label = None
            while label != current and label_path:
                # remove any labels that appeared outside of used sections
                label, used = label_path.popitem()
                if used and label != current:
                    # used label without end -> comment
                    comments.append(label)
            if used:
                # menu label was used - print menu label ending
                output.append(f"{line}\n")
            # find the new current (the last in list); if tree is empty return back to the root
            current = next(reversed(label_path.keys())) if label_path else None
        elif line == "#":
            # starting/ending possible menu label
            possibly_label = not possibly_label
        elif possibly_label:
            current = line[2:]  # remove leading '# '
            label_path[current] = False  # label not yet used
        elif line in min_options:
            # minimal config option detected
            for label, used in label_path.items():
                # print all menu labels that were not printed yet
                if not used:
                    output.append(f"\n#\n# {label}\n#\n")
                # mark the whole path from root as 'used'
                label_path[label] = True
            output.append(line + "\n")
    # Remove comments from minimal config, while keeping menu labels
    for comment in comments:
        output.remove(f"\n#\n# {comment}\n#\n")
    return "".join(output)


def write_min_config(deprecated_options, config, filename):
    idf_version = os.environ.get("IDF_VERSION", "")
    target_symbol = config.syms["IDF_TARGET"]
    # 'esp32` is harcoded here because the default value of IDF_TARGET is set on the first run from the environment
    # variable. I.E. `esp32  is not defined as default value.
    write_target = target_symbol.str_value != "esp32"

    CONFIG_HEADING = textwrap.dedent(
        f"""\
    # This file was generated using idf.py save-defconfig. It can be edited manually.
    # Espressif IoT Development Framework (ESP-IDF) {idf_version} Project Minimal Configuration
    #
    {target_symbol.config_string if write_target else ""}\
    """
    )

    if os.environ.get("ESP_IDF_KCONFIG_MIN_LABELS", False) == "1":
        lines = min_config_with_labels(config, CONFIG_HEADING).splitlines()
    else:
        lines = config._min_config_contents(header=CONFIG_HEADING).splitlines()

    # convert `# CONFIG_XY is not set` to `CONFIG_XY=n` to improve readability
    unset_match = re.compile(
        r"# {}([^ ]+) is not set".format(config.config_prefix)
    ).match
    for idx, line in enumerate(lines):
        match = unset_match(line)
        if match:
            lines[idx] = f"{config.config_prefix}{match.group(1)}=n"
    lines[-1] += "\n"
    config._write_if_changed(filename, "\n".join(lines))


def write_header(deprecated_options, config, filename):
    idf_version = os.environ.get("IDF_VERSION", "")
    CONFIG_HEADING = f"""/*
 * Automatically generated file. DO NOT EDIT.
 * Espressif IoT Development Framework (ESP-IDF) {idf_version} Configuration Header
 */
#pragma once
"""
    config.write_autoconf(filename, header=CONFIG_HEADING)
    deprecated_options.append_header(config, filename)


def write_cmake(deprecated_options, config, filename):
    with open(filename, "w") as f:
        tmp_dep_list = []
        write = f.write
        prefix = config.config_prefix

        write(
            """#
# Automatically generated file. DO NOT EDIT.
# Espressif IoT Development Framework (ESP-IDF) Configuration cmake include file
#
"""
        )

        configs_list = list()

        def write_node(node):
            sym = node.item
            if not isinstance(sym, kconfiglib.Symbol):
                return

            if sym.config_string:
                val = sym.str_value
                if (
                    sym.orig_type in (kconfiglib.BOOL, kconfiglib.TRISTATE)
                    and val == "n"
                ):
                    val = ""  # write unset values as empty variables
                elif sym.orig_type == kconfiglib.STRING:
                    val = kconfiglib.escape(val)
                elif sym.orig_type == kconfiglib.HEX:
                    val = hex(int(val, 16))  # ensure 0x prefix
                write('set({}{} "{}")\n'.format(prefix, sym.name, val))

                configs_list.append(prefix + sym.name)
                dep_opts = deprecated_options.get_deprecated_option(sym.name)
                for opt in dep_opts:
                    tmp_dep_list.append('set({}{} "{}")\n'.format(prefix, opt, val))
                    configs_list.append(prefix + opt)

        for n in config.node_iter():
            write_node(n)
        write("set(CONFIGS_LIST {})".format(";".join(configs_list)))

        if len(tmp_dep_list) > 0:
            write("\n# List of deprecated options for backward compatibility\n")
            f.writelines(tmp_dep_list)


def get_json_values(config):
    config_dict = {}

    def write_node(node):
        sym = node.item
        if not isinstance(sym, kconfiglib.Symbol):
            return

        if sym.config_string:
            val = sym.str_value
            if sym.type in [kconfiglib.BOOL, kconfiglib.TRISTATE]:
                val = val != "n"
            elif sym.type == kconfiglib.HEX:
                val = int(val, 16)
            elif sym.type == kconfiglib.INT:
                val = int(val)
            config_dict[sym.name] = val

    for n in config.node_iter(False):
        write_node(n)
    return config_dict


def write_json(deprecated_options, config, filename):
    config_dict = get_json_values(config)
    with open(filename, "w") as f:
        json.dump(config_dict, f, indent=4, sort_keys=True)


def get_menu_node_id(node):
    """Given a menu node, return a unique id
    which can be used to identify it in the menu structure

    Will either be the config symbol name, or a menu identifier
    'slug'

    """
    try:
        if not isinstance(node.item, kconfiglib.Choice):
            return node.item.name
    except AttributeError:
        pass

    result = []
    while node.parent is not None:
        slug = re.sub(r"\W+", "-", node.prompt[0]).lower()
        result.append(slug)
        node = node.parent

    result = "-".join(reversed(result))
    return result


def write_json_menus(deprecated_options, config, filename):
    existing_ids = set()
    result = []  # root level items
    node_lookup = {}  # lookup from MenuNode to an item in result

    def write_node(node):
        try:
            json_parent = node_lookup[node.parent]["children"]
        except KeyError:
            assert (
                node.parent not in node_lookup
            )  # if fails, we have a parent node with no "children" entity (ie a bug)
            json_parent = result  # root level node

        # node.kconfig.y means node has no dependency,
        if node.dep is node.kconfig.y:
            depends = None
        else:
            depends = kconfiglib.expr_str(node.dep)

        try:
            # node.is_menuconfig is True in newer kconfiglibs for menus and choices as well
            is_menuconfig = node.is_menuconfig and isinstance(
                node.item, kconfiglib.Symbol
            )
        except AttributeError:
            is_menuconfig = False

        new_json = None
        if node.item == kconfiglib.MENU or is_menuconfig:
            new_json = {
                "type": "menu",
                "title": node.prompt[0],
                "depends_on": depends,
                "children": [],
            }
            if is_menuconfig:
                sym = node.item
                new_json["name"] = sym.name
                new_json["help"] = node.help
                new_json["is_menuconfig"] = is_menuconfig
                greatest_range = None
                if len(sym.ranges) > 0:
                    # Note: Evaluating the condition using kconfiglib's expr_value
                    # should have one condition which is true
                    for min_range, max_range, cond_expr in sym.ranges:
                        if kconfiglib.expr_value(cond_expr):
                            greatest_range = [min_range, max_range]
                new_json["range"] = greatest_range

        elif isinstance(node.item, kconfiglib.Symbol):
            sym = node.item
            greatest_range = None
            if len(sym.ranges) > 0:
                # Note: Evaluating the condition using kconfiglib's expr_value
                # should have one condition which is true
                for min_range, max_range, cond_expr in sym.ranges:
                    if kconfiglib.expr_value(cond_expr):
                        base = 16 if sym.type == kconfiglib.HEX else 10
                        greatest_range = [
                            int(min_range.str_value, base),
                            int(max_range.str_value, base),
                        ]
                        break

            new_json = {
                "type": kconfiglib.TYPE_TO_STR[sym.type],
                "name": sym.name,
                "title": node.prompt[0] if node.prompt else None,
                "depends_on": depends,
                "help": node.help,
                "range": greatest_range,
                "children": [],
            }
        elif isinstance(node.item, kconfiglib.Choice):
            choice = node.item
            new_json = {
                "type": "choice",
                "title": node.prompt[0],
                "name": choice.name,
                "depends_on": depends,
                "help": node.help,
                "children": [],
            }

        if new_json:
            node_id = get_menu_node_id(node)
            if node_id in existing_ids:
                raise RuntimeError(
                    "Config file contains two items with the same id: %s (%s). "
                    % (node_id, node.prompt[0])
                    + "Please rename one of these items to avoid ambiguity."
                )
            new_json["id"] = node_id

            json_parent.append(new_json)
            node_lookup[node] = new_json

    for n in config.node_iter():
        write_node(n)
    with open(filename, "w") as f:
        f.write(json.dumps(result, sort_keys=True, indent=4))


def write_docs(deprecated_options, config, filename):
    try:
        target = os.environ["IDF_TARGET"]
    except KeyError:
        print("IDF_TARGET environment variable must be defined!")
        sys.exit(1)

    visibility = gen_kconfig_doc.ConfigTargetVisibility(config, target)
    gen_kconfig_doc.write_docs(config, visibility, filename)
    deprecated_options.append_doc(config, visibility, filename)


def update_if_changed(source, destination):
    with open(source, "r") as f:
        source_contents = f.read()

    if os.path.exists(destination):
        with open(destination, "r") as f:
            dest_contents = f.read()
        if source_contents == dest_contents:
            return  # nothing to update

    with open(destination, "w") as f:
        f.write(source_contents)


OUTPUT_FORMATS = {
    "config": write_config,
    "header": write_header,
    "cmake": write_cmake,
    "docs": write_docs,
    "json": write_json,
    "json_menus": write_json_menus,
    "savedefconfig": write_min_config,
}



#!/usr/bin/env python
#
# Long-running server process uses stdin & stdout to communicate JSON
# with a caller
#
import argparse
import json
import os
import sys
import tempfile

import kconfiglib

import kconfgen.core as kconfgen
from esp_idf_kconfig import __version__

# Min/Max supported protocol versions
MIN_PROTOCOL_VERSION = 1
MAX_PROTOCOL_VERSION = 2


def main():
    parser = argparse.ArgumentParser(
        description="kconfserver.py v%s - Config Generation Tool" % __version__,
        prog=os.path.basename(sys.argv[0]),
    )

    parser.add_argument(
        "--config", help="Project configuration settings", required=True
    )

    parser.add_argument(
        "--kconfig", help="KConfig file with config item definitions", required=True
    )

    parser.add_argument(
        "--sdkconfig-rename",
        help="File with deprecated Kconfig options",
        required=False,
    )

    parser.add_argument(
        "--env",
        action="append",
        default=[],
        help="Environment to set when evaluating the config file",
        metavar="NAME=VAL",
    )

    parser.add_argument(
        "--env-file",
        type=argparse.FileType("r"),
        help="Optional file to load environment variables from. Contents "
        "should be a JSON object where each key/value pair is a variable.",
    )

    parser.add_argument(
        "--version",
        help="Set protocol version to use on initial status",
        type=int,
        default=MAX_PROTOCOL_VERSION,
    )

    args = parser.parse_args()

    if args.version < MIN_PROTOCOL_VERSION:
        print(
            "Version %d is older than minimum supported protocol version %d. Client is much older than ESP-IDF version?"
            % (args.version, MIN_PROTOCOL_VERSION)
        )

    if args.version > MAX_PROTOCOL_VERSION:
        print(
            "Version %d is newer than maximum supported protocol version %d. Client is newer than ESP-IDF version?"
            % (args.version, MAX_PROTOCOL_VERSION)
        )

    try:
        args.env = [
            (name, value) for (name, value) in (e.split("=", 1) for e in args.env)
        ]
    except ValueError:
        print(
            "--env arguments must each contain =. To unset an environment variable, use 'ENV='"
        )
        sys.exit(1)

    for name, value in args.env:
        os.environ[name] = value

    if args.env_file is not None:
        env = json.load(args.env_file)
        os.environ.update(env)

    run_server(args.kconfig, args.config, args.sdkconfig_rename)


def run_server(
    kconfig, sdkconfig, sdkconfig_rename, default_version=MAX_PROTOCOL_VERSION
):
    config = kconfiglib.Kconfig(kconfig)
    sdkconfig_renames = [sdkconfig_rename] if sdkconfig_rename else []
    sdkconfig_renames_from_env = os.environ.get("COMPONENT_SDKCONFIG_RENAMES")
    if sdkconfig_renames_from_env:
        sdkconfig_renames += sdkconfig_renames_from_env.split(";")
    deprecated_options = kconfgen.DeprecatedOptions(
        config.config_prefix, path_rename_files=sdkconfig_renames
    )
    f_o = tempfile.NamedTemporaryFile(mode="w+b", delete=False)
    try:
        with open(sdkconfig, mode="rb") as f_i:
            f_o.write(f_i.read())
        f_o.close()  # need to close as DeprecatedOptions will reopen, and Windows only allows one open file
        deprecated_options.replace(sdkconfig_in=f_o.name, sdkconfig_out=sdkconfig)
    finally:
        os.unlink(f_o.name)
    config.load_config(sdkconfig)

    print("Server running, waiting for requests on stdin...", file=sys.stderr)

    config_dict = kconfgen.get_json_values(config)
    ranges_dict = get_ranges(config)
    visible_dict = get_visible(config)

    if default_version == 1:
        # V1: no 'visibility' key, send value None for any invisible item
        values_dict = dict(
            (k, v if visible_dict[k] else False) for (k, v) in config_dict.items()
        )
        json.dump(
            {"version": 1, "values": values_dict, "ranges": ranges_dict}, sys.stdout
        )
    else:
        # V2 onwards: separate visibility from version
        json.dump(
            {
                "version": default_version,
                "values": config_dict,
                "ranges": ranges_dict,
                "visible": visible_dict,
            },
            sys.stdout,
        )
    print("\n")
    sys.stdout.flush()

    while True:
        line = sys.stdin.readline()
        if not line:
            break
        try:
            req = json.loads(line)
        except ValueError as e:  # json module throws JSONDecodeError (sublcass of ValueError) on Py3 but ValueError on Py2
            response = {
                "version": default_version,
                "error": ["JSON formatting error: %s" % e],
            }
            json.dump(response, sys.stdout)
            print("\n")
            sys.stdout.flush()
            continue
        before = kconfgen.get_json_values(config)
        before_ranges = get_ranges(config)
        before_visible = get_visible(config)

        if "load" in req:  # load a new sdkconfig

            if req.get("version", default_version) == 1:
                # for V1 protocol, send all items when loading new sdkconfig.
                # (V2+ will only send changes, same as when setting an item)
                before = {}
                before_ranges = {}
                before_visible = {}

            # if no new filename is supplied, use existing sdkconfig path, otherwise update the path
            if req["load"] is None:
                req["load"] = sdkconfig
            else:
                sdkconfig = req["load"]

        if "save" in req:
            if req["save"] is None:
                req["save"] = sdkconfig
            else:
                sdkconfig = req["save"]

        error = handle_request(deprecated_options, config, req)

        after = kconfgen.get_json_values(config)
        after_ranges = get_ranges(config)
        after_visible = get_visible(config)

        values_diff = diff(before, after)
        ranges_diff = diff(before_ranges, after_ranges)
        visible_diff = diff(before_visible, after_visible)
        if req["version"] == 1:
            # V1 response, invisible items have value None
            for k in (k for (k, v) in visible_diff.items() if not v):
                values_diff[k] = None
            response = {"version": 1, "values": values_diff, "ranges": ranges_diff}
        else:
            # V2+ response, separate visibility values
            response = {
                "version": req["version"],
                "values": values_diff,
                "ranges": ranges_diff,
                "visible": visible_diff,
            }
        if error:
            for err in error:
                print("Error: %s" % err, file=sys.stderr)
            response["error"] = error
        json.dump(response, sys.stdout)
        print("\n")
        sys.stdout.flush()


def handle_request(deprecated_options, config, req):
    if "version" not in req:
        return ["All requests must have a 'version'"]

    if req["version"] < MIN_PROTOCOL_VERSION or req["version"] > MAX_PROTOCOL_VERSION:
        return [
            "Unsupported request version %d. Server supports versions %d-%d"
            % (req["version"], MIN_PROTOCOL_VERSION, MAX_PROTOCOL_VERSION)
        ]

    error = []

    if "load" in req:
        print("Loading config from %s..." % req["load"], file=sys.stderr)
        try:
            config.load_config(req["load"])
        except Exception as e:
            error += ["Failed to load from %s: %s" % (req["load"], e)]

    if "set" in req:
        handle_set(config, error, req["set"])

    if "save" in req:
        try:
            print("Saving config to %s..." % req["save"], file=sys.stderr)
            kconfgen.write_config(deprecated_options, config, req["save"])
        except Exception as e:
            error += ["Failed to save to %s: %s" % (req["save"], e)]

    return error


def handle_set(config, error, to_set):
    missing = [k for k in to_set if k not in config.syms]
    if missing:
        error.append(
            "The following config symbol(s) were not found: %s" % (", ".join(missing))
        )
    # replace name keys with the full config symbol for each key:
    to_set = dict((config.syms[k], v) for (k, v) in to_set.items() if k not in missing)

    # Work through the list of values to set, noting that
    # some may not be immediately applicable (maybe they depend
    # on another value which is being set). Therefore, defer
    # knowing if any value is unsettable until then end

    while len(to_set):
        set_pass = [(k, v) for (k, v) in to_set.items() if k.visibility]
        if not set_pass:
            break  # no visible keys left
        for (sym, val) in set_pass:
            if sym.type in (kconfiglib.BOOL, kconfiglib.TRISTATE):
                if val is True:
                    sym.set_value(2)
                elif val is False:
                    sym.set_value(0)
                else:
                    error.append(
                        "Boolean symbol %s only accepts true/false values" % sym.name
                    )
            elif sym.type == kconfiglib.HEX:
                try:
                    if not isinstance(val, int):
                        val = int(
                            val, 16
                        )  # input can be a decimal JSON value or a string of hex digits
                    sym.set_value(hex(val))
                except ValueError:
                    error.append(
                        "Hex symbol %s can accept a decimal integer or a string of hex digits, only"
                    )
            else:
                sym.set_value(str(val))
            print("Set %s" % sym.name)
            del to_set[sym]

    if len(to_set):
        error.append(
            "The following config symbol(s) were not visible so were not updated: %s"
            % (", ".join(s.name for s in to_set))
        )


def diff(before, after):
    """
    Return a dictionary with the difference between 'before' and 'after',
    for items which are present in 'after' dictionary
    """
    diff = dict((k, v) for (k, v) in after.items() if before.get(k, None) != v)
    return diff


def get_ranges(config):
    ranges_dict = {}

    def is_base_n(i, n):
        try:
            int(i, n)
            return True
        except ValueError:
            return False

    def get_active_range(sym):
        """
        Returns a tuple of (low, high) integer values if a range
        limit is active for this symbol, or (None, None) if no range
        limit exists.
        """
        base = (
            kconfiglib._TYPE_TO_BASE[sym.orig_type]
            if sym.orig_type in kconfiglib._TYPE_TO_BASE
            else 0
        )

        try:
            for low_expr, high_expr, cond in sym.ranges:
                if kconfiglib.expr_value(cond):
                    low = (
                        int(low_expr.str_value, base)
                        if is_base_n(low_expr.str_value, base)
                        else 0
                    )
                    high = (
                        int(high_expr.str_value, base)
                        if is_base_n(high_expr.str_value, base)
                        else 0
                    )
                    return (low, high)
        except ValueError:
            pass
        return (None, None)

    def handle_node(node):
        sym = node.item
        if not isinstance(sym, kconfiglib.Symbol):
            return
        active_range = get_active_range(sym)
        if active_range[0] is not None:
            ranges_dict[sym.name] = active_range

    for n in config.node_iter():
        handle_node(n)
    return ranges_dict


def get_visible(config):
    """
    Return a dict mapping node IDs (config names or menu node IDs) to True/False for their visibility
    """
    result = {}
    menus = []

    # when walking the menu the first time, only
    # record whether the config symbols are visible
    # and make a list of menu nodes (that are not symbols)
    def handle_node(node):
        sym = node.item
        try:
            visible = sym.visibility != 0
            result[node] = visible
        except AttributeError:
            menus.append(node)

    for n in config.node_iter():
        handle_node(n)

    # now, figure out visibility for each menu. A menu is visible if any of its children are visible
    for m in reversed(menus):  # reverse to start at leaf nodes
        result[m] = any(v for (n, v) in result.items() if n.parent == m)

    # return a dict mapping the node ID to its visibility.
    result = dict((kconfgen.get_menu_node_id(n), v) for (n, v) in result.items())

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# gen_kconfig_doc - kconfgen.py support for generating ReST markup documentation
#
# For each option in the loaded Kconfig (e.g. 'FOO'), CONFIG_FOO link target is
# generated, allowing options to be referenced in other documents
# (using :ref:`CONFIG_FOO`)
#
# SPDX-FileCopyrightText: 2017-2022 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
from __future__ import print_function

import re

import kconfiglib

# Indentation to be used in the generated file
INDENT = "    "

# Characters used when underlining section heading
HEADING_SYMBOLS = '#*=-^"+'

# Keep the heading level in sync with api-reference/kconfig.rst
INITIAL_HEADING_LEVEL = 3
MAX_HEADING_LEVEL = len(HEADING_SYMBOLS) - 1


class ConfigTargetVisibility(object):
    """
    Determine the visibility of Kconfig options based on IDF targets. Note that other environment variables should not
    imply invisibility and neither dependencies on visible options with default disabled state. This difference makes
    it necessary to implement our own visibility and cannot use the visibility defined inside Kconfiglib.
    """

    def __init__(self, config, target):
        # target actually is not necessary here because kconfiglib.expr_value() will evaluate it internally
        self.config = config
        self.visibility = (
            dict()
        )  # node name to (x, y) mapping where x is the visibility (True/False) and y is the
        # name of the config which implies the visibility
        self.target_env_var = "IDF_TARGET"
        self.direct_eval_set = frozenset(
            [
                kconfiglib.EQUAL,
                kconfiglib.UNEQUAL,
                kconfiglib.LESS,
                kconfiglib.LESS_EQUAL,
                kconfiglib.GREATER,
                kconfiglib.GREATER_EQUAL,
            ]
        )

    def _implies_invisibility(self, item):
        if isinstance(item, tuple):
            if item[0] == kconfiglib.NOT:
                (invisibility, source) = self._implies_invisibility(item[1])
                if source is not None and source.startswith(self.target_env_var):
                    return (not invisibility, source)
                else:
                    # we want to be visible all configs which are not dependent on target variables,
                    # e.g. "depends on XY" and "depends on !XY" as well
                    return (False, None)
            elif item[0] == kconfiglib.AND:
                (invisibility, source) = self._implies_invisibility(item[1])
                if invisibility:
                    return (True, source)
                (invisibility, source) = self._implies_invisibility(item[2])
                if invisibility:
                    return (True, source)
                return (False, None)
            elif item[0] == kconfiglib.OR:
                implication_list = [
                    self._implies_invisibility(item[1]),
                    self._implies_invisibility(item[2]),
                ]
                if all([implies for (implies, _) in implication_list]):
                    source_list = [
                        s
                        for (_, s) in implication_list
                        if s.startswith(self.target_env_var)
                    ]
                    # if source_list has more items then it should not matter which will imply the invisibility
                    return (True, source_list[0])
                return (False, None)
            elif item[0] in self.direct_eval_set:

                def node_is_invisible(item):
                    return all([node.prompt is None for node in item.nodes])

                if node_is_invisible(item[1]) or node_is_invisible(item[1]):
                    # it makes no sense to call self._implies_invisibility() here because it won't generate any useful
                    # "source"
                    return (not kconfiglib.expr_value(item), None)
                else:
                    # expressions with visible configs can be changed to make the item visible
                    return (False, None)
            else:
                raise RuntimeError("Unimplemented operation in {}".format(item))
        else:  # Symbol or Choice
            vis_list = [self._visible(node) for node in item.nodes]
            if len(vis_list) > 0 and all([not visible for (visible, _) in vis_list]):
                source_list = [
                    s
                    for (_, s) in vis_list
                    if s is not None and s.startswith(self.target_env_var)
                ]
                # if source_list has more items then it should not matter which will imply the invisibility
                return (True, source_list[0])

            if item.name.startswith(self.target_env_var):
                return (not kconfiglib.expr_value(item), item.name)

            if len(vis_list) == 1:
                (visible, source) = vis_list[0]
                if visible:
                    return (
                        False,
                        item.name,
                    )  # item.name is important here in case the result will be inverted: if
                    # the dependency is on another config then it can be still visible

            return (False, None)

    def _visible(self, node):
        if node.item == kconfiglib.COMMENT:
            return (False, None)
        if isinstance(node.item, kconfiglib.Symbol) or isinstance(
            node.item, kconfiglib.Choice
        ):
            dependencies = node.item.direct_dep  # "depends on" for configs
            name_id = node.item.name
            simple_def = len(node.item.nodes) <= 1  # defined only in one source file
            # Probably it is not necessary to check the default statements.
        else:
            dependencies = node.visibility  # "visible if" for menu
            name_id = node.prompt[0]
            simple_def = False  # menus can be defined with the same name at multiple locations and they don't know
            # about each other like configs through node.item.nodes. Therefore, they cannot be stored and have to be
            # re-evaluated always.

        try:
            (visib, source) = self.visibility[name_id]
        except KeyError:

            def invert_first_arg(_tuple):
                return (not _tuple[0], _tuple[1])

            (visib, source) = (
                self._visible(node.parent) if node.parent else (True, None)
            )

            if visib:
                (visib, source) = invert_first_arg(
                    self._implies_invisibility(dependencies)
                )

            if simple_def:
                # Configs defined at multiple places are not stored because they could have different visibility based
                # on different targets. kconfiglib.expr_value() will handle the visibility.
                self.visibility[name_id] = (visib, source)

        return (
            visib,
            source,
        )  # not used in "finally" block because failure messages from _implies_invisibility are
        # this way more understandable

    def visible(self, node):
        if not node.prompt:
            # don't store this in self.visibility because don't want to stop at invisible nodes when recursively
            # searching for invisible targets
            return False

        return self._visible(node)[0]


def write_docs(config, visibility, filename):
    """Note: writing .rst documentation ignores the current value
    of any items. ie the --config option can be ignored.
    (However at time of writing it still needs to be set to something...)"""
    with open(filename, "w") as f:
        for node in config.node_iter():
            write_menu_item(f, node, visibility)


def node_is_menu(node):
    try:
        return node.item in [kconfiglib.MENU, kconfiglib.COMMENT] or node.is_menuconfig
    except AttributeError:
        return False  # not all MenuNodes have is_menuconfig for some reason


def get_breadcrumbs(node):
    # this is a bit wasteful as it recalculates each time, but still...
    result = []
    node = node.parent
    while node.parent:
        if node.prompt:
            result = [":ref:`%s`" % get_link_anchor(node)] + result
        node = node.parent
    return " > ".join(result)


def get_link_anchor(node):
    try:
        return "CONFIG_%s" % node.item.name
    except AttributeError:
        assert node_is_menu(node)  # only menus should have no item.name

    # for menus, build a link anchor out of the parents
    result = []
    while node.parent:
        if node.prompt:
            result = [re.sub(r"[^a-zA-z0-9]+", "-", node.prompt[0])] + result
        node = node.parent
    result = "-".join(result).lower()
    return result


def get_heading_level(node):
    result = INITIAL_HEADING_LEVEL
    node = node.parent
    while node.parent:
        result += 1
        if result == MAX_HEADING_LEVEL:
            return MAX_HEADING_LEVEL
        node = node.parent
    return result


def format_rest_text(text, indent):
    # Format an indented text block for use with ReST
    text = indent + text.replace("\n", "\n" + indent)
    # Escape some characters which are inline formatting in ReST
    text = text.replace("*", "\\*")
    text = text.replace("_", "\\_")
    # replace absolute links to documentation by relative ones
    text = re.sub(
        r"https://docs.espressif.com/projects/esp-idf/\w+/\w+/(.+)\.html",
        r":doc:`../\1`",
        text,
    )
    text += "\n"
    return text


def _minimize_expr(expr, visibility):
    def expr_nodes_invisible(e):
        return (
            hasattr(e, "nodes")
            and len(e.nodes) > 0
            and all(not visibility.visible(i) for i in e.nodes)
        )

    if isinstance(expr, tuple):
        if expr[0] == kconfiglib.NOT:
            new_expr = _minimize_expr(expr[1], visibility)
            return (
                kconfiglib.Kconfig.y
                if new_expr == kconfiglib.Kconfig.n
                else kconfiglib.Kconfig.n
            )
        else:
            new_expr1 = _minimize_expr(expr[1], visibility)
            new_expr2 = _minimize_expr(expr[2], visibility)
            if expr[0] == kconfiglib.AND:
                if (
                    new_expr1 == kconfiglib.Kconfig.n
                    or new_expr2 == kconfiglib.Kconfig.n
                ):
                    return kconfiglib.Kconfig.n
                if new_expr1 == kconfiglib.Kconfig.y:
                    return new_expr2
                if new_expr2 == kconfiglib.Kconfig.y:
                    return new_expr1
            elif expr[0] == kconfiglib.OR:
                if (
                    new_expr1 == kconfiglib.Kconfig.y
                    or new_expr2 == kconfiglib.Kconfig.y
                ):
                    return kconfiglib.Kconfig.y
                if new_expr1 == kconfiglib.Kconfig.n:
                    return new_expr2
                if new_expr2 == kconfiglib.Kconfig.n:
                    return new_expr1
            elif expr[0] == kconfiglib.EQUAL:
                if not isinstance(new_expr1, type(new_expr2)):
                    return kconfiglib.Kconfig.n
                if new_expr1 == new_expr2:
                    return kconfiglib.Kconfig.y
            elif expr[0] == kconfiglib.UNEQUAL:
                if not isinstance(new_expr1, type(new_expr2)):
                    return kconfiglib.Kconfig.y
                if new_expr1 != new_expr2:
                    return kconfiglib.Kconfig.n
            else:  # <, <=, >, >=
                if not isinstance(new_expr1, type(new_expr2)):
                    return kconfiglib.Kconfig.n  # e.g "True < 2"

                if expr_nodes_invisible(new_expr1) or expr_nodes_invisible(new_expr2):
                    return (
                        kconfiglib.Kconfig.y
                        if kconfiglib.expr_value(expr)
                        else kconfiglib.Kconfig.n
                    )

            return (expr[0], new_expr1, new_expr2)

    if (
        not kconfiglib.expr_value(expr)
        and len(expr.config_string) == 0
        and expr_nodes_invisible(expr)
    ):
        # nodes which are invisible
        # len(expr.nodes) > 0 avoids constant symbols without actual node definitions, e.g. integer constants
        # len(expr.config_string) == 0 avoids hidden configs which reflects the values of choices
        return kconfiglib.Kconfig.n

    if (
        kconfiglib.expr_value(expr)
        and len(expr.config_string) > 0
        and expr_nodes_invisible(expr)
    ):
        # hidden config dependencies which will be written to sdkconfig as enabled ones.
        return kconfiglib.Kconfig.y

    if any(node.item.name.startswith(visibility.target_env_var) for node in expr.nodes):
        # We know the actual values for IDF_TARGETs
        return (
            kconfiglib.Kconfig.y
            if kconfiglib.expr_value(expr)
            else kconfiglib.Kconfig.n
        )

    return expr


def write_menu_item(f, node, visibility):
    def is_choice(node):
        """Skip choice nodes, they are handled as part of the parent (see below)"""
        return isinstance(node.parent.item, kconfiglib.Choice)

    if is_choice(node) or not visibility.visible(node):
        return

    try:
        name = node.item.name
    except AttributeError:
        name = None

    is_menu = node_is_menu(node)

    # Heading
    if name:
        title = "CONFIG_%s" % name
    else:
        # if no symbol name, use the prompt as the heading
        title = node.prompt[0]

    f.write(".. _%s:\n\n" % get_link_anchor(node))
    f.write("%s\n" % title)
    f.write(HEADING_SYMBOLS[get_heading_level(node)] * len(title))
    f.write("\n\n")

    if name:
        f.write("%s%s\n\n" % (INDENT, node.prompt[0]))
        f.write("%s:emphasis:`Found in:` %s\n\n" % (INDENT, get_breadcrumbs(node)))

    try:
        if node.help:
            # Help text normally contains newlines, but spaces at the beginning of
            # each line are stripped by kconfiglib. We need to re-indent the text
            # to produce valid ReST.
            f.write(format_rest_text(node.help, INDENT))
            f.write("\n")
    except AttributeError:
        pass  # No help

    if isinstance(node.item, kconfiglib.Choice):
        f.write("%sAvailable options:\n\n" % INDENT)
        choice_node = node.list
        while choice_node:
            # Format available options as a list
            # First, link anchor for this option
            f.write("%s  .. _%s:\n\n" % (INDENT * 2, get_link_anchor(choice_node)))
            # Then, option itself, as a list item
            f.write(
                "%s- %-20s (%s%s)\n"
                % (
                    INDENT * 2,
                    choice_node.prompt[0],
                    node.kconfig.config_prefix,
                    choice_node.item.name,
                )
            )
            if choice_node.help:
                HELP_INDENT = INDENT * 2
                fmt_help = format_rest_text(choice_node.help, "  " + HELP_INDENT)
                f.write("%s  \n%s\n" % (HELP_INDENT, fmt_help))
            choice_node = choice_node.next
            f.write("\n")

        f.write("\n\n")

    if isinstance(node.item, kconfiglib.Symbol):

        def _expr_str(sc):
            if sc.is_constant or not sc.nodes:
                return "{}".format(sc.name)
            opt_name = "%s%s" % (sc.kconfig.config_prefix, sc.name)
            if sc.choice:
                # link targets not associated with a section cannot be referenced without providing the title
                # https://github.com/sphinx-doc/sphinx/issues/9993
                return ":ref:`%s<%s>`" % (opt_name, opt_name)
            return ":ref:`%s`" % opt_name

        range_strs = []
        for low, high, cond in node.item.ranges:
            cond = _minimize_expr(cond, visibility)
            if cond == kconfiglib.Kconfig.n:
                continue
            if not isinstance(cond, tuple) and cond != kconfiglib.Kconfig.y:
                if len(cond.nodes) > 0 and all(
                    not visibility.visible(i) for i in cond.nodes
                ):
                    if not kconfiglib.expr_value(cond):
                        continue
            range_str = "%s- from %s to %s" % (
                INDENT * 2,
                low.str_value,
                high.str_value,
            )
            if cond != kconfiglib.Kconfig.y and not kconfiglib.expr_value(cond):
                range_str += " if %s" % kconfiglib.expr_str(cond, _expr_str)
            range_strs.append(range_str)
        if len(range_strs) > 0:
            f.write("%sRange:\n" % INDENT)
            f.write("\n".join(range_strs))
            f.write("\n\n")

        default_strs = []
        for default, cond in node.item.defaults:
            cond = _minimize_expr(cond, visibility)
            if cond == kconfiglib.Kconfig.n:
                continue
            if not isinstance(cond, tuple) and cond != kconfiglib.Kconfig.y:
                if len(cond.nodes) > 0 and all(
                    not visibility.visible(i) for i in cond.nodes
                ):
                    if not kconfiglib.expr_value(cond):
                        continue
            # default.type is mostly UNKNOWN so it cannot be used reliably for detecting the type
            d = default.str_value
            if d in ["y", "Y"]:
                d = "Yes (enabled)"
            elif d in ["n", "N"]:
                d = "No (disabled)"
            elif re.search(
                r"[^0-9a-fA-F]", d
            ):  # simple string detection: if it not a valid number
                d = '"%s"' % d
            default_str = "%s- %s" % (INDENT * 2, d)
            if cond != kconfiglib.Kconfig.y and not kconfiglib.expr_value(cond):
                default_str += " if %s" % kconfiglib.expr_str(cond, _expr_str)
            default_strs.append(default_str)
        if len(default_strs) > 0:
            f.write("%sDefault value:\n" % INDENT)
            f.write("\n".join(default_strs))
            f.write("\n\n")

    if is_menu:
        # enumerate links to child items
        child_list = []
        child = node.list
        while child:
            if not is_choice(child) and child.prompt and visibility.visible(child):
                child_list.append((child.prompt[0], get_link_anchor(child)))
            child = child.next
        if len(child_list) > 0:
            f.write("Contains:\n\n")
            sorted_child_list = sorted(child_list, key=lambda pair: pair[0].lower())
            ref_list = ["- :ref:`{}`".format(anchor) for _, anchor in sorted_child_list]
            f.write("\n".join(ref_list))
            f.write("\n\n")



#!/usr/bin/env python
#
# Command line tool to check kconfig files
#
# SPDX-FileCopyrightText: 2018-2023 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import argparse
import os
import re

# ouput file with suggestions will get this suffix
OUTPUT_SUFFIX = ".new"

SPACES_PER_INDENT = 4

CONFIG_NAME_MAX_LENGTH = 40

CONFIG_NAME_MIN_PREFIX_LENGTH = 3

# The checker will not fail if it encounters this string (it can be used for temporarily resolve conflicts)
RE_NOERROR = re.compile(r"\s+#\s+NOERROR\s+$")

# list or rules for lines
LINE_ERROR_RULES = [
    # (regular expression for finding,      error message,                                  correction)
    (
        re.compile(r"\t"),
        "tabulators should be replaced by spaces",
        r" " * SPACES_PER_INDENT,
    ),
    (re.compile(r"\s+\n"), "trailing whitespaces should be removed", r"\n"),
    (re.compile(r".{120}"), "line should be shorter than 120 characters", None),
]


class InputError(RuntimeError):
    """
    Represents and error on the input
    """

    def __init__(self, path, line_number, error_msg, suggested_line):
        super(InputError, self).__init__(
            "{}:{}: {}".format(path, line_number, error_msg)
        )
        self.suggested_line = suggested_line


class BaseChecker(object):
    """
    Base class for all checker objects
    """

    def __init__(self, path_in_idf):
        self.path_in_idf = path_in_idf

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass


class SourceChecker(BaseChecker):
    # allow to source only files which will be also checked by the script
    # Note: The rules are complex and the LineRuleChecker cannot be used
    def process_line(self, line, line_number):
        m = re.search(r'^\s*[ro]{0,2}source(\s*)"([^"]+)"', line)

        if m:
            if len(m.group(1)) == 0:
                raise InputError(
                    self.path_in_idf,
                    line_number,
                    '"source" has to been followed by space',
                    line.replace("source", "source "),
                )
            path = m.group(2)
            filename = os.path.basename(path)
            if path in [
                "$COMPONENT_KCONFIGS_SOURCE_FILE",
                "$COMPONENT_KCONFIGS_PROJBUILD_SOURCE_FILE",
            ]:
                pass
            elif not filename.startswith("Kconfig."):
                raise InputError(
                    self.path_in_idf,
                    line_number,
                    "only filenames starting with Kconfig.* can be sourced",
                    line.replace(
                        path, os.path.join(os.path.dirname(path), "Kconfig." + filename)
                    ),
                )


class LineRuleChecker(BaseChecker):
    """
    checks LINE_ERROR_RULES for each line
    """

    def process_line(self, line, line_number):
        suppress_errors = RE_NOERROR.search(line) is not None
        errors = []
        for rule in LINE_ERROR_RULES:
            m = rule[0].search(line)
            if m:
                if suppress_errors:
                    # just print but no failure
                    e = InputError(self.path_in_idf, line_number, rule[1], line)
                    print(e)
                else:
                    errors.append(rule[1])
                if rule[2]:
                    line = rule[0].sub(rule[2], line)
        if len(errors) > 0:
            raise InputError(self.path_in_idf, line_number, "; ".join(errors), line)


class IndentAndNameChecker(BaseChecker):
    """
    checks the indentation of each line and configuration names
    """

    def __init__(self, path_in_idf, debug=False):
        super(IndentAndNameChecker, self).__init__(path_in_idf)
        self.debug = debug
        self.min_prefix_length = CONFIG_NAME_MIN_PREFIX_LENGTH

        # stack of the nested menuconfig items, e.g. ['mainmenu', 'menu', 'config']
        self.level_stack = []

        # stack common prefixes of configs
        self.prefix_stack = []

        # if the line ends with '\' then we force the indent of the next line
        self.force_next_indent = 0

        # menu items which increase the indentation of the next line
        self.re_increase_level = re.compile(
            r"""^\s*
                                          (
                                               (menu(?!config))
                                              |(mainmenu)
                                              |(choice)
                                              |(config)
                                              |(menuconfig)
                                              |(help)
                                              |(if)
                                              |(source)
                                              |(osource)
                                              |(rsource)
                                              |(orsource)
                                          )
                                       """,
            re.X,
        )

        # closing menu items which decrease the indentation
        self.re_decrease_level = re.compile(
            r"""^\s*
                                          (
                                               (endmenu)
                                              |(endchoice)
                                              |(endif)
                                          )
                                       """,
            re.X,
        )

        # matching beginning of the closing menuitems
        self.pair_dic = {
            "endmenu": "menu",
            "endchoice": "choice",
            "endif": "if",
        }

        # regex for config names
        self.re_name = re.compile(
            r"""^
                                       (
                                            (?:config)
                                           |(?:menuconfig)
                                           |(?:choice)

                                       )\s+
                                       (\w+)
                                      """,
            re.X,
        )

        # regex for new prefix stack
        self.re_new_stack = re.compile(
            r"""^
                                            (
                                                 (?:menu(?!config))
                                                |(?:mainmenu)
                                                |(?:choice)

                                            )
                                            """,
            re.X,
        )

    def __exit__(self, type, value, traceback):
        super(IndentAndNameChecker, self).__exit__(type, value, traceback)
        if len(self.prefix_stack) > 0:
            self.check_common_prefix("", "EOF")
        if len(self.prefix_stack) != 0:
            if self.debug:
                print(self.prefix_stack)
            raise RuntimeError(
                "Prefix stack should be empty. Perhaps a menu/choice hasn't been closed"
            )

    def del_from_level_stack(self, count):
        """delete count items from the end of the level_stack"""
        if count > 0:
            # del self.level_stack[-0:] would delete everything and we expect not to delete anything for count=0
            del self.level_stack[-count:]

    def update_level_for_inc_pattern(self, new_item):
        if self.debug:
            print("level+", new_item, ": ", self.level_stack, end=" -> ")
        # "config" and "menuconfig" don't have a closing pair. So if new_item is an item which need to be indented
        # outside the last "config" or "menuconfig" then we need to find to a parent where it belongs
        if new_item in [
            "config",
            "menuconfig",
            "menu",
            "choice",
            "if",
            "source",
            "rsource",
            "osource",
            "orsource",
        ]:
            # item is not belonging to a previous "config" or "menuconfig" so need to indent to parent
            for i, item in enumerate(reversed(self.level_stack)):
                if item in ["menu", "mainmenu", "choice", "if"]:
                    # delete items ("config", "menuconfig", "help") until the appropriate parent
                    self.del_from_level_stack(i)
                    break
            else:
                # delete everything when configs are at top level without a parent menu, mainmenu...
                self.del_from_level_stack(len(self.level_stack))

        self.level_stack.append(new_item)
        if self.debug:
            print(self.level_stack)
        # The new indent is for the next line. Use the old one for the current line:
        return len(self.level_stack) - 1

    def update_level_for_dec_pattern(self, new_item):
        if self.debug:
            print("level-", new_item, ": ", self.level_stack, end=" -> ")
        target = self.pair_dic[new_item]
        for i, item in enumerate(reversed(self.level_stack)):
            # find the matching beginning for the closing item in reverse-order search
            # Note: "menuconfig", "config" and "help" don't have closing pairs and they are also on the stack. Now they
            # will be deleted together with the "menu" or "choice" we are closing.
            if item == target:
                i += 1  # delete also the matching beginning
                if self.debug:
                    print("delete ", i, end=" -> ")
                self.del_from_level_stack(i)
                break
        if self.debug:
            print(self.level_stack)
        return len(self.level_stack)

    def check_name_and_update_prefix(self, line, line_number):
        m = self.re_name.search(line)
        if m:
            name = m.group(2)
            name_length = len(name)

            if name_length > CONFIG_NAME_MAX_LENGTH:
                raise InputError(
                    self.path_in_idf,
                    line_number,
                    "{} is {} characters long and it should be {} at most"
                    "".format(name, name_length, CONFIG_NAME_MAX_LENGTH),
                    line + "\n",
                )  # no suggested correction for this
            if len(self.prefix_stack) == 0:
                self.prefix_stack.append(name)
            elif self.prefix_stack[-1] is None:
                self.prefix_stack[-1] = name
            else:
                # this has nothing common with paths but the algorithm can be used for this also
                self.prefix_stack[-1] = os.path.commonprefix(
                    [self.prefix_stack[-1], name]
                )
            if self.debug:
                print("prefix+", self.prefix_stack)
        m = self.re_new_stack.search(line)
        if m:
            self.prefix_stack.append(None)
            if self.debug:
                print("prefix+", self.prefix_stack)

    def check_common_prefix(self, line, line_number):
        common_prefix = self.prefix_stack.pop()
        if self.debug:
            print("prefix-", self.prefix_stack)
        if common_prefix is None:
            return
        common_prefix_len = len(common_prefix)
        if common_prefix_len < self.min_prefix_length:
            raise InputError(
                self.path_in_idf,
                line_number,
                'The common prefix for the config names of the menu ending at this line is "{}".\n'
                "\tAll config names in this menu should start with the same prefix of {} characters "
                "or more.".format(common_prefix, self.min_prefix_length),
                line,
            )  # no suggested correction for this
        if len(self.prefix_stack) > 0:
            parent_prefix = self.prefix_stack[-1]
            if parent_prefix is None:
                # propagate to parent level where it will influence the prefix checking with the rest which might
                # follow later on that level
                self.prefix_stack[-1] = common_prefix
            else:
                if len(self.level_stack) > 0 and self.level_stack[-1] in [
                    "mainmenu",
                    "menu",
                ]:
                    # the prefix from menu is not required to propagate to the children
                    return
                if not common_prefix.startswith(parent_prefix):
                    raise InputError(
                        self.path_in_idf,
                        line_number,
                        'Common prefix "{}" should start with {}'
                        "".format(common_prefix, parent_prefix),
                        line,
                    )  # no suggested correction for this

    def process_line(self, line, line_number):
        stripped_line = line.strip()
        if len(stripped_line) == 0:
            self.force_next_indent = 0
            return
        current_level = len(self.level_stack)
        m = re.search(r"\S", line)  # indent found as the first non-space character
        if m:
            current_indent = m.start()
        else:
            current_indent = 0

        if current_level > 0 and self.level_stack[-1] == "help":
            if current_indent >= current_level * SPACES_PER_INDENT:
                # this line belongs to 'help'
                self.force_next_indent = 0
                return

        if self.force_next_indent > 0:
            if current_indent != self.force_next_indent:
                raise InputError(
                    self.path_in_idf,
                    line_number,
                    "Indentation consists of {} spaces instead of {}".format(
                        current_indent, self.force_next_indent
                    ),
                    (" " * self.force_next_indent) + line.lstrip(),
                )
            else:
                if not stripped_line.endswith("\\"):
                    self.force_next_indent = 0
                return

        elif stripped_line.endswith("\\") and stripped_line.startswith(
            ("config", "menuconfig", "choice")
        ):
            raise InputError(
                self.path_in_idf,
                line_number,
                "Line-wrap with backslash is not supported here",
                line,
            )  # no suggestion for this

        self.check_name_and_update_prefix(stripped_line, line_number)

        m = self.re_increase_level.search(line)
        if m:
            current_level = self.update_level_for_inc_pattern(m.group(1))
        else:
            m = self.re_decrease_level.search(line)
            if m:
                new_item = m.group(1)
                current_level = self.update_level_for_dec_pattern(new_item)
                if new_item not in ["endif"]:
                    # endif doesn't require to check the prefix because the items inside if/endif belong to the
                    # same prefix level
                    self.check_common_prefix(line, line_number)

        expected_indent = current_level * SPACES_PER_INDENT

        if stripped_line.endswith("\\"):
            self.force_next_indent = expected_indent + SPACES_PER_INDENT
        else:
            self.force_next_indent = 0

        if current_indent != expected_indent:
            raise InputError(
                self.path_in_idf,
                line_number,
                "Indentation consists of {} spaces instead of {}".format(
                    current_indent, expected_indent
                ),
                (" " * expected_indent) + line.lstrip(),
            )


def valid_directory(path):
    if not os.path.isdir(path):
        raise argparse.ArgumentTypeError("{} is not a valid directory!".format(path))
    return path


def validate_kconfig_file(
    kconfig_full_path, verbose=False
):  # type: (str, bool) -> bool
    suggestions_full_path = kconfig_full_path + OUTPUT_SUFFIX
    fail = False

    with open(kconfig_full_path, "r", encoding="utf-8") as f, open(
        suggestions_full_path, "w", encoding="utf-8", newline="\n"
    ) as f_o, LineRuleChecker(kconfig_full_path) as line_checker, SourceChecker(
        kconfig_full_path
    ) as source_checker, IndentAndNameChecker(
        kconfig_full_path, debug=verbose
    ) as indent_and_name_checker:
        try:
            for line_number, line in enumerate(f, start=1):
                try:
                    for checker in [
                        line_checker,
                        indent_and_name_checker,
                        source_checker,
                    ]:
                        checker.process_line(line, line_number)
                    # The line is correct therefore we echo it to the output file
                    f_o.write(line)
                except InputError as e:
                    print(e)
                    fail = True
                    f_o.write(e.suggested_line)
        except UnicodeDecodeError:
            raise ValueError(
                "The encoding of {} is not Unicode.".format(kconfig_full_path)
            )

    if fail:
        print(
            "\t{} has been saved with suggestions for resolving the issues.\n"
            "\tPlease note that the suggestions can be wrong and "
            "you might need to re-run the checker several times "
            "for solving all issues".format(suggestions_full_path)
        )
        print(
            "\tPlease fix the errors and run {} for checking the correctness of "
            "Kconfig files.".format(os.path.abspath(__file__))
        )
        return False
    else:
        print("{}: OK".format(kconfig_full_path))
        try:
            os.remove(suggestions_full_path)
        except Exception:
            # not a serious error is when the file cannot be deleted
            print("{} cannot be deleted!".format(suggestions_full_path))
        finally:
            return True


def main():
    parser = argparse.ArgumentParser(description="Kconfig style checker")
    parser.add_argument(
        "files",
        nargs="*",
        help="Kconfig files to check (full paths separated by space)",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Print more information (useful for debugging)",
    )
    args = parser.parse_args()

    success_counter = 0
    failure_counter = 0

    files = [os.path.abspath(file_path) for file_path in args.files]

    for full_path in files:
        is_valid = validate_kconfig_file(full_path, args.verbose)
        if is_valid:
            success_counter += 1
        else:
            failure_counter += 1

    if success_counter > 0:
        print("{} files have been successfully checked.".format(success_counter))
    if failure_counter > 0:
        print(
            "{} files have errors. Please take a look at the log.".format(
                failure_counter
            )
        )
        return 1

    if not files:
        print("WARNING: no files specified.")
    return 0
